(* Definition of the kernel of an homomorphism. *)
(* Venanzio Capretta, January 1999              *)
(* Coq version 6.2.3                            *)

Require Export Homomorphisms.
Require Export Congruences.

Implicit Arguments On.

Section Kernels.

Variable  sigma : Signature.
Variables A, B  : (Algebra sigma).
Variable  phi   : (Homomorphism A B).

Local n := (sorts_num sigma).
Local m := (fun_num   sigma).

Local As := (sorts A).
Local Bs := (sorts B).

Local ker_rel : (s : (Finite n))(relation (As s)) :=
  [s,a1,a2] ((phi s) a1) [=] ((phi s) a2).

Remark ker_well_defined :
  (s:(Finite n))(Relation_well_defined (ker_rel 1!s)).

Proof.

Intro s.
Red.
Intros x1 x2 y1 y2 xeq yeq.
Unfold ker_rel.
Cut ((phi s) x2) [=] ((phi s) x1).
Intros H1.
Cut ((phi s) y1) [=] ((phi s) y2).
Intros H2.
Intro H.
EApply s_transitivity.
Apply H1.

EApply s_transitivity.
Apply H.

Apply H2.

Apply (s_fun_proof (phi s)).
Assumption.

Apply (s_fun_proof (phi s)).
Apply s_symmetry.
Assumption.

Qed.

Transparent ker_well_defined.

Local ker_setoid : (s : (Finite n))(Setoid_relation (As s)) :=
  [s](setoid_relation (ker_well_defined 1!s)).

Remark ker_reflexive : (s : (Finite n))(reflexive (As s) (ker_rel 1!s)).

Proof.

Intros s.
Red.
Intro x.
Red.
Apply s_reflexivity.

Qed.

Transparent ker_reflexive.

Remark ker_symmetric : (s : (Finite n))(symmetric (As s) (ker_rel 1!s)).

Proof.

Intros s.
Red.
Intros x y.
Unfold ker_rel.
Intro H.
Apply s_symmetry.
Assumption.

Qed.

Transparent ker_symmetric.

Remark ker_transitive : (s : (Finite n))(transitive (As s) (ker_rel 1!s)).

Proof.

Intros s.
Red.
Intros x y z.
Unfold ker_rel.
Intros H1 H2.
EApply s_transitivity.
Apply H1.
Apply H2.

Qed.

Transparent ker_transitive.

Remark ker_equivalence : (s : (Finite n))(Equiv (ker_rel 1!s)).

Proof.

Intro s.
Red.
Split.
Exact (ker_reflexive 1!s).

Split.
Exact (ker_transitive 1!s).

Exact (ker_symmetric 1!s).

Qed.

Transparent ker_equivalence.

Section Kernel_Substitutivity.

(* We prove the substitutivity property for the kernel. *)

Variable i : (Finite m).

Local h := (nth_arity   i).
Local a := (nth_arg   2!i).
Local r := (nth_result  i).

Local fiA := (functions A i).
Local fiB := (functions B i).

Variables args1A, args2A : (Fun_arg_arguments A i).
Variable  argsAeq : (j : (Finite h))((ker_setoid (a j)) (args1A j) (args2A j)).

Local args1B : (Fun_arg_arguments B i) := [j]((phi (a j)) (args1A j)).
Local args2B : (Fun_arg_arguments B i) := [j]((phi (a j)) (args2A j)).

Remark argsBeq : (j : (Finite h)) (args1B j) [=] (args2B j).
Proof. Exact argsAeq. Qed.

Transparent argsBeq.

Remark argseq : args1B [=] args2B.

Proof.

(Simpl; Red).
Exact argsBeq.

Qed.

Transparent argseq.

Remark phi_eq : (fiB args1B) [=] (fiB args2B).

Proof.

Apply (s_fun_proof fiB).
Exact argseq.

Qed.

Transparent phi_eq.

Remark phi_comm1 : (fiB args1B) [=] ((phi r) (fiA args1A)).
Proof. Apply (hom_proof phi args1A). Qed.

Transparent phi_comm1.

Remark phi_comm2 : (fiB args2B) [=] ((phi r) (fiA args2A)).
Proof. Apply (hom_proof phi args2A). Qed.

Transparent phi_comm2.

Lemma ker_subst : ((ker_setoid r) (fiA args1A) (fiA args2A)).

Proof.

Simpl.
Unfold ker_rel.
EApply s_transitivity.
Apply s_symmetry.
Apply phi_comm1.

EApply s_transitivity.
Apply phi_eq.

Apply phi_comm2.

Qed.

End Kernel_Substitutivity.


Remark ker_substitutivity : (is_Congruence ker_setoid).
Proof. Exact ker_subst. Qed.

Transparent ker_substitutivity.

Definition kernel : (Congruence A) :=
  (congruence 3!ker_setoid ker_equivalence ker_substitutivity).

Definition ker_quot : (Algebra sigma) := (quotient kernel).

Local kers := (sorts ker_quot).

(* Natural map from A to the quotient A/kernel. *)

Local nat_fun : (s : (Finite n))(As s) -> (kers s) := [s,a]a.

Remark nat_fun_well_defined :
  (s : (Finite n))(fun_well_defined (nat_fun 1!s)).

Proof.

Intros s.
Red.
Intros a1 a2 aeq.
Simpl.
Unfold nat_fun.
Apply (!sr_proof (As s) (kernel s) a1 a1 a1 a2).
Apply s_reflexivity.

Assumption.

Apply (cong_reflexivity kernel a1).

Qed.

Transparent nat_fun_well_defined.

Local nat_s_fun : (s : (Finite n))(S_Fun (As s) (kers s)) :=
  [s](setoid_function (nat_fun_well_defined 1!s)).

(* We prove now that nat_s_fun is an homomorphism. *)

Section Nat_Homomorphism.

Variable i : (Finite m).

Local h := (nth_arity    i).
Local a := (nth_arg    2!i).
Local r := (nth_result   i).

Local fiA := (functions A i).
Local fik := (functions ker_quot i).

Variable args : (Fun_arg_arguments A i).

Local argsk : (Fun_arg_arguments ker_quot i) :=
  [j]((nat_s_fun (a j)) (args j)).

Remark args_eq : (s_eq 1!(Fun_arg_arguments A i) argsk args).

Proof.

Simpl.
Red.
Intros i'.
Apply s_reflexivity.

Qed.

Transparent args_eq.

Local resA  := (fiA args).
Local resk  := (fik argsk).

Remark resAk_eq : (s_eq 1!(sorts A r) resk resA).
(* in this case it is necessary to specify the setoid otherwise
   it will be assumed that the equality is in the setoid of the
   first element (and not the second as we want.
*)

Proof.
Apply (s_fun_proof fiA 4!argsk 5!args).
Exact args_eq.
Qed.

Transparent resAk_eq.

Local resA' := ((nat_s_fun r) resA).

Remark nat_commutes_A : (s_eq 1!(sorts A r) resk resA').

Proof.

EApply s_transitivity.
Apply resAk_eq.

Apply s_symmetry.
Apply s_reflexivity.

Qed.

Transparent nat_commutes_A.

Lemma nat_commutes : resk [=] resA'.
(* Notice that this statement take as granted that the equality
   is the one in the setoid of the first element: (kers r).
*)

Proof.

Simpl.
Apply (!sr_proof (As r) (ker_setoid r) resk resk resk resA').
Apply s_reflexivity.

Exact nat_commutes_A.

Simpl.
Apply (cong_reflexivity kernel resk).

Qed.

End Nat_Homomorphism.

Remark nat_is_homomorphism : (Is_homomorphism 2!A 3!ker_quot nat_s_fun).
Proof. Exact nat_commutes. Qed.

Transparent nat_is_homomorphism.

Definition natural : (Homomorphism A ker_quot) :=
  (homomorphism nat_is_homomorphism).

End Kernels.

Implicit Arguments Off.
